package org.apache.activemq.test;

import java.util.Date;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.DeliveryMode;
import javax.jms.Destination;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import junit.framework.Assert;

/*
import org.apache.activemq.ActiveMQConnectionFactory;
import org.apache.activemq.broker.BrokerService;
import org.apache.activemq.broker.region.policy.PolicyEntry;
import org.apache.activemq.broker.region.policy.PolicyMap;
import org.apache.activemq.broker.jmx.ManagementContext;
import org.apache.activemq.command.ActiveMQQueue;

import org.apache.qpid.jms.example.Sender;
*/
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Unit test for AMQ-XXXX.
 * Qpid JMS clients fail to send a larger number of messages to the ActiveMQ
 * broker using local JMS transactions. The producer hangs after around 50 
 * messages.
 * 
 * This unit test cannot use an embedded broker. It requires an external
 * broker to listen for AMQP on port 5672.
 * Broker URL and connection settings are configured in 
 * src/main/resources/jndi.properties
 * 
 * The embedded broker needs to include activemq-amqp, which includes
 * org.apache.qpid:proton-jms:jar:0.8. That jar does not work with 
 * org.apache.qpid:qpid-jms-examples:jar:0.3.0 which is used by the 
 * AMQP producer in this test. 
 * Having both on the classpath results in Connection.start() to hang 
 * in the Producer code below.
 *
 */
public class AMQPTransactionTest {

    // protected static final String BROKER_URI = "amqp://localhost:5672";
    // private static final String QUEUE_NAME= "AMQP-TEST";
    // private static final ActiveMQQueue QUEUE = new ActiveMQQueue(QUEUE_NAME);
    // private BrokerService broker;

    private static final int MESSAGE_SIZE = 1024;
    public static final int NUM_MSGS = 500;
    private static final Logger LOG = LoggerFactory.getLogger(AMQPTransactionTest.class);
    private CountDownLatch producerLatch = new CountDownLatch(1);
    public boolean createBroker = false;

    /**
     * Creates a new ActiveMQConnectionFactory.
     * @return ConnectionFactory
     * @throws Exception
    protected ConnectionFactory createConnectionFactory() throws Exception {
        ActiveMQConnectionFactory connectionFactory = new ActiveMQConnectionFactory(BROKER_URI);
        return connectionFactory;
    }
    */


    @Before
    public void setUp() throws Exception {
        /* cannot use embedded broker
        if(createBroker) {
            broker = createBroker();
            broker.start();
        }
        factory = createConnectionFactory();
        */
    }


    @After
    public void tearDown() throws Exception {
        /* cannot use embedded broker
        if(createBroker) {
            broker.stop();
        }
        */
    }


    /**
     * Uses qpid-jms-client library to send NUM_MSGS to broker.
     * Due to bug AMQ-XXXX only 50 messages get sent, then the producer
     * hangs after some messages.
     * 
     * @throws Exception any exceptions are propagated.
     */
    @Test(timeout=60*1000)
    public void testAMQPWithTransactions() throws Exception {
        final AtomicLong sendCounter = new AtomicLong();

        // kick off a producer that pre-fills the queue
        QPidProducer producer = new QPidProducer(sendCounter);
        producer.start();
        if (!producerLatch.await(20, TimeUnit.SECONDS)) {
            LOG.error("Test failed and timed out after 20 seconds");
        }
        if (!producer.success) {
            LOG.error("Test failed to send {} messages.", NUM_MSGS);
            Assert.fail("Test failed to send " + NUM_MSGS + " messages.");
        }
    }


    /**
     * Creates an embedded broker instance
     * @return the BrokerService instance
     * @throws Exception 
    protected BrokerService createBroker() throws Exception {
        BrokerService broker = new BrokerService();
        broker.setUseJmx(true);

        ManagementContext mgtCtx = new ManagementContext();
        mgtCtx.setConnectorPort(1100);
        broker.setManagementContext(mgtCtx);
        broker.getManagementContext().setCreateConnector(true);

        broker.setPersistent(false);
        broker.getSystemUsage().getMemoryUsage().setLimit(1024*1024*120);

        broker.addConnector(BROKER_URI + "?transport.trace=true");
        return broker;
    }
    */


    /**
     * Sample message producer implemented as its own thread.
     * Sends NUM_MSGS to destination QUEUE
     */
    private final class QPidProducer extends Thread {
        public boolean success = false;
        private final AtomicLong sendCounter;

        private QPidProducer(AtomicLong sendCounter) {
            this.sendCounter = sendCounter;
        }

        @Override
        public void run() {
            Connection conn=null;
            try {

                // The configuration for the Qpid InitialContextFactory has been supplied in
                // a jndi.properties file in the classpath, which results in it being picked
                // up automatically by the InitialContext constructor.
                Context context = new InitialContext();

                ConnectionFactory factory = (ConnectionFactory) context.lookup("myFactoryLookup");
                Destination queue = (Destination) context.lookup("myQueueLookup");
                conn = factory.createConnection("admin", "admin");
                conn.setExceptionListener(new MyExceptionListener());
                
                // Connection.start hangs if org.apache.qpid:proton-jms:jar:0.8
                // and org.apache.qpid:qpid-jms-examples:jar:0.3.0 are both on
                // the classpath.
                conn.start();
                Session session = conn.createSession(true, Session.SESSION_TRANSACTED);

                MessageProducer producer = session.createProducer(queue);
                producer.setDeliveryMode(DeliveryMode.PERSISTENT);

                int i=0;
                while(sendCounter.get() < NUM_MSGS) {
                    // producer.send(session.createTextMessage(createMessageText(i++)));
                    producer.send(session.createTextMessage(createMessageText(i++)), DeliveryMode.PERSISTENT, Message.DEFAULT_PRIORITY, Message.DEFAULT_TIME_TO_LIVE);
                    session.commit();
                    sendCounter.incrementAndGet();
                    LOG.info("Sent message #{}", sendCounter.get());
                }
                success=true;
            } catch (NamingException ne) {
                LOG.error("Could not resolve ConnectionFactory.", ne);
            } catch (JMSException e) {
                LOG.error("Failed to send messages. ", e);
            } finally {
                try {
                    conn.close();
                    producerLatch.countDown();
                } catch (Throwable e) {
                    LOG.error(e.getMessage());
                }
            }
        }

        /* it seems org.apache.qpid:qpid-jms-examples:jar:0.3.0
           is not found on any repos. So not using this Sender class
           which does the same as above run() method.
        @Override
        public void run() {
            Connection conn=null;
            try {

            Sender sender = new Sender();
            String[] args = {};
            sender.main(args);
            } catch (Exception e) {
                e.printStackTrace();
            } finally {
                    producerLatch.countDown();
            }
        }
        */
        /**
         * Creates a String of a given size filling with 
         * space characters.
         * @param index
         * @return
         */
        private String createMessageText(int index) {
            StringBuffer buffer = new StringBuffer(MESSAGE_SIZE);
            buffer.append(index + " on " + new Date() + " ...");
            if (buffer.length() > MESSAGE_SIZE) {
                return buffer.substring(0, MESSAGE_SIZE);
            }
            for (int i = buffer.length(); i < MESSAGE_SIZE; i++) {
                buffer.append(' ');
            }
            return buffer.toString();
        }
    }

    private static class MyExceptionListener implements ExceptionListener {
        @Override
        public void onException(JMSException exception) {
            System.out.println("Connection ExceptionListener fired, exiting.");
            exception.printStackTrace(System.out);
            System.exit(1);
        }
    }
}
